﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.Bandits {
  // helper to create canonical forms of expressions
  // TODO: change symbolicregressionpoly10problem to use this class
  // this does not support expressions with constants (in transformations we assume constant opt is used)
  // (e.g. we transform all negative signs to + because it is assumed a negative constant can be produced for the term)
  public class ExpressionExtender {

    // supports the grammar
    // G(E):
    //     E -> V | V+E | V-E | V*E | V%E | (E) 
    //     V -> <variables>
    //     ";

    // transformations: 
    // - transform c*b*a to a*b*c (lexicographic ordering for factors, cached)
    // - transfrom - to + (assumed optimized constants for terms)


    // most-recently-used caching (with limited capacity) for canonical representations
    MostRecentlyUsedCache<string, string> canonicalPhraseCache = new MostRecentlyUsedCache<string, string>(100000);
    public string CanonicalRepresentation(string phrase) {
      int pos = 0;
      return CanonicalExpr(phrase, ref pos);
    }

    private Expr CanonicalExpr(string phrase, ref int pos) {
      var terms = new List<Term>();
      terms.Add(CanonicalTerm(phrase, ref pos));
      var curSy = phrase[pos];
      while (curSy == '+' || curSy == '-' || curSy == '^') {
        if (curSy == '+') {
          pos++;
          terms.Add(CanonicalTerm(phrase, ref pos));
        } else if (curSy == '-') {
          pos++;
          terms.Add(CanonicalTerm(phrase, ref pos)); // minus is the same as plus assuming constant opt
        } else {
          pos++;
          throw new NotImplementedException();
          // var e = Expr(variables, constants);
          // r = Not(r) * e + r * Not(e); // xor = (!x AND y) OR (x AND !y)
        }
        curSy = phrase[pos];
      }

      // at this point we might have multiple constants in the list of terms but we only need one of them
      var firstConstant = terms.OfType<Variable>().FirstOrDefault(); // if a term is only a variable then it must be a constant (otherwise (for real variables) it would be a product of two constants)
      if (firstConstant == null) {
        // no constant found => add a constant offset for each full expression
        var offset = new Variable();
        constants.Add(offset);
        terms.Add(offset);
      } else {
        // there is already a constant => remove all others if there are more
        var otherConsts = terms.OfType<Variable>().Where(t => t != firstConstant).ToArray();
        foreach (var otherConst in otherConsts) {
          terms.Remove(otherConst);
          constants.Remove(otherConst);
        }
      }
      return TermBuilder.Sum(terms);
    }

    private Term CanonicalTerm(string phrase, ref int pos) {
      var factors = new List<Factor>();
      var invFactors = new List<Factor>();
      var f = CanonicalFact(phrase, ref pos);
      if (f != null) factors.Add(f);
      var curSy = phrase[pos];
      while (curSy == '*' || curSy == '%') {
        if (curSy == '*') {
          pos++;
          f = CanonicalFact(phrase, ref pos); // fact might return string.Empty (for constants)
          if (f != null) factors.Add(f);
        } else {
          pos++;
          f = CanonicalFact(phrase, ref pos);
          if (f != null) invFactors.Add(f);
        }
        curSy = phrase[pos];
      }

      // cancel factors if they are contained in factors and invFactors
      foreach (var cancelledF in factors.Intersect(invFactors).ToArray()) {
        factors.Remove(cancelledF);
        invFactors.Remove(cancelledF);
      }
      // at this point factors and invFactors might be empty
      if (factors.Count == 0 && invFactors.Count == 0) throw new NotImplementedException();

      if(factors.Count == 1)
      // extend factors if they are composed of terms
      // this means that we return a list of terms instead of one single term
      // at this point factors can only be a list of additive terms
      var extendedFactors = new List<Factor>();

      // order factors (bysize then lexicographically)
    }

    private Factor CanonicalFact(string phrase, ref int pos) {
      var curSy = phrase[pos];
      if (curSy == '!') {
        throw new NotSupportedException();
      } else if (curSy == '(') {
        pos++;
        Expr r = CanonicalExpr(phrase, ref pos);
        if (phrase[pos] != ')') throw new ArgumentException();
        pos++;
        if (r.Kind == ExprKind.SimpleExpr)
          return new Factor(r.Term);
        else
          return new Factor(r.Terms); // a composite factor
      } else if (curSy >= 'a' && curSy <= 'z') {
        pos++;
        return new Factor(curSy.ToString());
        //       } else if (curSy >= '0' && curSy <= '9') {
      } else throw new ArgumentException("found symbol " + curSy);
    }

    #region factor
    private enum FactorKind { SimpleFactor, Expr }
    private class Factor {
      public readonly FactorKind Kind;
      public readonly string Value; // only set for Kind == SimpleFactor
      public readonly IEnumerable<string> Values; // only set for Kind == Expr 

      public Factor(string f) {
        this.Kind = FactorKind.SimpleFactor;
        this.Value = f;
      }
      public Factor(IEnumerable<string> factorTerms) {
        this.Kind = FactorKind.Expr;
        this.Values = factorTerms;
      }

      public override string ToString() {
        if (Kind == FactorKind.SimpleFactor) return Value;
        else return string.Join("+", Values);
      }
      public override bool Equals(object obj) {
        var other = obj as Factor;
        if (other == null) return false;
        if (other.Kind != this.Kind) return false;
        if (Kind == FactorKind.SimpleFactor) return this.Value == other.Value;
        if (this.Values.Count() != other.Values.Count()) return false;
        if (this.Values.Zip(other.Values, Tuple.Create).All(t => t.Item1 == t.Item2)) return true;
        return false;
      }
      public override int GetHashCode() {
        var h = 31415;
        if (Value != null) h ^= Value.GetHashCode();
        foreach (var v in Values) {
          h ^= v.GetHashCode();
        }
        return h;
      }
    }
    #endregion
    #region term
    private enum TermKind { SimpleTerm, Term }
    private class Term {
      public readonly TermKind Kind;
      public readonly Factor Factor; // only set for Kind == Simple
      public readonly IEnumerable<Factor> Factors;

      public Term(Factor f) {
        this.Kind = TermKind.SimpleTerm;
        this.Factor = f;
      }
      public Term(IEnumerable<Factor> factors) {
        this.Kind = TermKind.Term;
        this.Factors = factors;
      }

      public override string ToString() {
        if (Kind == TermKind.SimpleTerm) return Factor.ToString();
        else return string.Join("+", Factors);
      }
      public override bool Equals(object obj) {
        var other = obj as Term;
        if (other == null) return false;
        if (other.Kind != this.Kind) return false;
        if (Kind == TermKind.SimpleTerm) return this.Factor == other.Factor;
        if (this.Factors.Count() != other.Factors.Count()) return false;
        if (this.Factors.Zip(other.Factors, Tuple.Create).All(t => t.Item1 == t.Item2)) return true;
        return false;
      }
      public override int GetHashCode() {
        var h = 31415;
        if (Factor != null) h ^= Factor.GetHashCode();
        foreach (var t in Factors) {
          h ^= t.GetHashCode();
        }
        return h;
      }
    }
    #endregion
    #region expr
    private enum ExprKind { SimpleExpr, Expr }
    private class Expr {
      public readonly ExprKind Kind;
      public readonly string Term; // only set for Kind == SimpleExpr
      public readonly IEnumerable<string> Terms; // only set for Kind == Expr 

      public Expr(string f) {
        this.Kind = ExprKind.SimpleExpr;
        this.Term = f;
      }
      public Expr(IEnumerable<string> exprTerms) {
        this.Kind = ExprKind.Expr;
        this.Terms = exprTerms;
      }

      public override string ToString() {
        if (Kind == ExprKind.SimpleExpr) return Term;
        else return string.Join("+", Terms);
      }
      public override bool Equals(object obj) {
        var other = obj as Expr;
        if (other == null) return false;
        if (other.Kind != this.Kind) return false;
        if (Kind == ExprKind.SimpleExpr) return this.Term == other.Term;
        if (this.Terms.Count() != other.Terms.Count()) return false;
        if (this.Terms.Zip(other.Terms, Tuple.Create).All(t => t.Item1 == t.Item2)) return true;
        return false;
      }
      public override int GetHashCode() {
        var h = 31415;
        if (Term != null) h ^= Term.GetHashCode();
        foreach (var t in Terms) {
          h ^= t.GetHashCode();
        }
        return h;
      }
    }
    #endregion
    // private enum TermKind { }
    // private class Term {
    //   public TermKind Kind;
    //   public string FirstFactor; // always set
    //   public IEnumerable<string> RemainingFactors; // only set for Kind == Expr 
    // }

    // cache the canonical form of terms for performance reasons
    private Dictionary<string, string> canonicalTermDictionary = new Dictionary<string, string>();
    private string CanonicalTerm(string term) {
      string canonicalTerm;
      if (!canonicalTermDictionary.TryGetValue(term, out canonicalTerm)) {
        // add
        var chars = term.ToCharArray();
        Array.Sort(chars);
        var sb = new StringBuilder(chars.Length);
        // we want to have the up-case characters last
        for (int i = chars.Length - 1; i > 0; i--) {
          if (chars[i] != '*') {
            sb.Append(chars[i]);
            if (chars[i - 1] != '*') sb.Append('*');
          }
        }
        if (chars[0] != '*') sb.Append(chars[0]); // last term
        canonicalTerm = sb.ToString();
        canonicalTermDictionary.Add(term, canonicalTerm);
      }
      return canonicalTerm;
    }
  }
}
